import asyncio
import json
import logging
import sys
from pathlib import Path
from traceback import FrameSummary, StackSummary
from typing import Annotated, Any

import typer

from icij_worker import TaskState
from icij_worker.cli.utils import AsyncTyper, eprint
from icij_worker.http_ import TaskClient
from icij_worker.objects import READY_STATES, Task, TaskError

logger = logging.getLogger(__name__)

DEFAULT_SERVICE_ADDRESS = "http://localhost:8000"

_ARGS_HELP = "task argument as a JSON string or file path"
_GROUP_HELP = "task group"
_SERVICE_URL_HELP = "service URL address"
_POLLING_INTERVAL_S_HELP = "task state polling interval in seconds"
_NAME_HELP = "registered task name"
_RESULT_HELP = "get a task result"
_START_HELP = "creates a new task and start it"
_TASK_ID_HELP = "task ID"
_WATCH_HELP = "watch a task until it's complete"

TaskArgs = str

task_app = AsyncTyper(name="tasks")


@task_app.command(help=_START_HELP)
async def start(
    name: Annotated[str, typer.Argument(help=_NAME_HELP)],
    args: Annotated[TaskArgs, typer.Argument(help=_ARGS_HELP)] = None,
    group: Annotated[
        str | None,
        typer.Option("--group", "-g", help=_GROUP_HELP),
    ] = None,
    service_address: Annotated[
        str, typer.Option("--ds-address", "-a", help=_SERVICE_URL_HELP)
    ] = DEFAULT_SERVICE_ADDRESS,
):
    match args:
        case str():
            as_path = Path(name)
            if as_path.exists():
                args = json.loads(as_path.read_text())
            else:
                args = json.loads(args)
        case None:
            args = dict()
        case _:
            raise TypeError(f"Invalid args {args}")
    client = TaskClient(service_address)
    async with client:
        task_id = await client.create_task(name, args, group=group)
    eprint(f"Task({task_id}) started !")
    eprint(f"Task({task_id}) 🛫")
    print(task_id)


@task_app.command(help=_WATCH_HELP)
async def watch(
    task_id: Annotated[str, typer.Argument(help=_TASK_ID_HELP)],
    service_address: Annotated[
        str, typer.Option("--ds-address", "-a", help=_SERVICE_URL_HELP)
    ] = DEFAULT_SERVICE_ADDRESS,
    polling_interval_s: Annotated[
        float, typer.Option("--polling-interval-s", "-p", help=_POLLING_INTERVAL_S_HELP)
    ] = 1.0,
):
    client = TaskClient(service_address)
    async with client:
        task = await client.get_task(task_id)
        if task.state is READY_STATES:
            await _handle_ready(task, client, already_done=True)
        await _handle_alive(task, client, polling_interval_s)
    print(task_id)


@task_app.command(help=_RESULT_HELP)
async def result(
    task_id: Annotated[str, typer.Argument(help=_TASK_ID_HELP)],
    service_address: Annotated[
        str, typer.Option("--ds-address", "-a", help=_SERVICE_URL_HELP)
    ] = DEFAULT_SERVICE_ADDRESS,
) -> Any:
    client = TaskClient(service_address)
    async with client:
        res = await client.get_task_result(task_id)
        if isinstance(res, (dict, list)):
            res = json.dumps(res, indent=2)
        print(res)


async def _handle_ready(
    task: Task, client: TaskClient, already_done: bool = False
) -> None:
    match task.state:
        case TaskState.ERROR:
            await _handle_error(task, client)
        case TaskState.CANCELLED:
            await _handle_cancelled(task)
        case TaskState.DONE:
            if already_done:
                await _handle_already_done(task)
            else:
                await _handle_done(task)
        case _:
            raise ValueError(f"Unexpected task state {task.state}")


async def _handle_error(task, client: TaskClient):
    error = await client.get_task_error(task.id)
    eprint(
        f"Task({task.id}) failed with the following"
        f" error:\n\n{_format_error(error)}"
    )
    eprint(f"Task({task.id}) ❌")
    raise typer.Exit(code=1)


async def _handle_cancelled(task):
    eprint(f"Task({task.id}) was cancelled !")
    eprint(f"Task({task.id}) 🛑")
    raise typer.Exit(code=1)


async def _handle_already_done(task):
    eprint(f"Task({task.id}) ✅ is already completed !")


async def _handle_done(task):
    eprint(f"Task({task.id}) 🛬")
    eprint(f"Task({task.id}) ✅")


async def _handle_alive(
    task: Task, client: TaskClient, polling_interval_s: float
) -> None:
    from alive_progress import alive_bar

    title = f"Task({task.id}) 🛫"
    stats = "(ETA: {eta})"
    monitor = "{percent}"
    progress_bar = alive_bar(
        title=title, manual=True, stats=stats, monitor=monitor, file=sys.stderr
    )
    with progress_bar as bar:
        task_state = task.state
        while task_state not in READY_STATES:
            task = await client.get_task(task.id)
            task_state = task.state
            progress = task.progress or 0.0
            bar(progress)  # pylint: disable=not-callable
            await asyncio.sleep(polling_interval_s)
    if task_state in READY_STATES:
        await _handle_ready(task, client)


def _format_error(error: TaskError) -> str:
    stack = StackSummary.from_list(
        [FrameSummary(f.name, f.lineno, f.name) for f in error.stacktrace]
    )
    msg = f"{error.name}:\n{stack}\n{error.message}"
    if error.cause:
        msg += "\n cause by {error.cause}"
    return msg



import importlib.metadata
import os
from typing import Annotated

import typer

import icij_worker
from icij_common.logging_utils import setup_loggers
from icij_worker.cli.tasks import task_app
from icij_worker.cli.utils import AsyncTyper
from icij_worker.cli.workers import worker_app

cli_app = AsyncTyper(context_settings={"help_option_names": ["-h", "--help"]})
cli_app.add_typer(worker_app)
cli_app.add_typer(task_app)


def version_callback(value: bool):
    if value:
        package_version = importlib.metadata.version(icij_worker.__name__)
        print(package_version)
        raise typer.Exit()


def pretty_exc_callback(value: bool):
    if not value:
        os.environ["_TYPER_STANDARD_TRACEBACK"] = "1"


@cli_app.callback(name="icij-worker")
def main(
    version: Annotated[  # pylint: disable=unused-argument
        bool | None,
        typer.Option(  # pylint: disable=unused-argument
            "--version", callback=version_callback, is_eager=True
        ),
    ] = None,
    pretty_exceptions: Annotated[  # pylint: disable=unused-argument
        bool,
        typer.Option(  # pylint: disable=unused-argument
            "--pretty-exceptions", callback=pretty_exc_callback, is_eager=True
        ),
    ] = False,
):
    """Python async worker pool CLI 🧑‍🏭"""
    setup_loggers(["__main__", icij_worker.__name__])



import asyncio
import logging
from abc import ABC, abstractmethod
from asyncio import Task as AsyncIOTask
from functools import cached_property
from typing import ClassVar, final

from pydantic import Field

from icij_common.pydantic_utils import merge_configs, no_enum_values_config, safe_copy
from icij_common.registrable import RegistrableConfig, RegistrableFromConfig
from icij_worker import AsyncApp, ResultEvent, Task, TaskState
from icij_worker.app import AsyncAppConfig
from icij_worker.exceptions import TaskAlreadyQueued, UnknownTask, UnregisteredTask
from icij_worker.objects import (
    AsyncBackend,
    CancelledEvent,
    ErrorEvent,
    ManagerEvent,
    ProgressEvent,
)
from icij_worker.routing_strategy import RoutingStrategy
from icij_worker.task_storage import TaskStorage

logger = logging.getLogger(__name__)


class TaskManagerConfig(RegistrableConfig):
    model_config = merge_configs(no_enum_values_config())

    registry_key: ClassVar[str] = Field(frozen=True, default="backend")
    backend: AsyncBackend = Field(frozen=True)

    app_path: str
    app_config: AsyncAppConfig = Field(default_factory=AsyncAppConfig)

    @property
    def app(self) -> AsyncApp:
        app = AsyncApp.load(self.app_path).with_config(self.app_config)
        return app


class TaskManager(TaskStorage, RegistrableFromConfig, ABC):
    def __init__(self, app: AsyncApp):
        self._app = app
        self._loop = asyncio.get_event_loop()
        self._consume_loop: AsyncIOTask | None = None

    @final
    async def __aenter__(self):
        await self._aenter__()
        self._consume_loop = self._loop.create_task(self.consume_events())

    async def _aenter__(self):
        pass

    @final
    async def __aexit__(self, exc_type, exc_value, tb):
        await self._aexit__(exc_type, exc_value, tb)
        if self._consume_loop is not None and not self._consume_loop.done():
            logger.info("cancelling worker event loop...")
            self._consume_loop.cancel()
            await asyncio.wait([self._consume_loop])
            logger.info("worker event loop cancelled")

    async def _aexit__(self, exc_type, exc_val, exc_tb):
        pass

    @cached_property
    def late_ack(self) -> bool:
        return self._app.config.late_ack

    @cached_property
    def max_task_queue_size(self) -> int:
        return self._app.config.max_task_queue_size

    @cached_property
    def _routing_strategy(self) -> RoutingStrategy:
        return self._app.routing_strategy

    @cached_property
    def app_name(self) -> str:
        return self._app.name

    @final
    async def enqueue(self, task: Task) -> Task:
        if task.state is not TaskState.CREATED:
            msg = f"invalid state {task.state}, expected {TaskState.CREATED}"
            raise ValueError(msg)
        task = await self.get_task(task.id)
        if task.state is TaskState.QUEUED:
            raise TaskAlreadyQueued(task.id)
        await self._enqueue(task)
        queued = safe_copy(task, update={"state": TaskState.QUEUED})
        await self.save_task(queued)
        return queued

    @final
    async def requeue(self, task: Task):
        logger.info("requeing Task(id=%s)", task.id)
        update = {"state": TaskState.QUEUED, "progress": 0.0}
        updated = safe_copy(task, update=update)
        await self._requeue(updated)
        logger.info("Task(id=%s) requeued", updated.id)

    @final
    async def save_task(self, task: Task) -> bool:
        max_retries = None
        try:
            group = await self.get_task_group(task_id=task.id)
        except UnknownTask as e:
            try:
                group = self._app.registry[task.name].group
                if group is not None:
                    group = group.name
                max_retries = self._app.registry[task.name].max_retries
            except KeyError:
                available_tasks = list(self._app.registry)
                raise UnregisteredTask(task.name, available_tasks) from e
        task = task.with_max_retries(max_retries)
        return await self.save_task_(task, group)

    async def _save_cancelled_event(self, event: CancelledEvent):
        task = await self.get_task(event.task_id)
        task = task.as_resolved(event)
        if event.requeue and not self.late_ack:
            await self.requeue(task)
        await self.save_task(task)

    @final
    async def consume_events(self):
        while True:
            msg = await self._consume()
            if isinstance(msg, ResultEvent):
                logger.debug("saving result for task: %s", msg.task_id)
                await self._save_result_event(msg)
            elif isinstance(msg, ErrorEvent):
                logger.debug("saving error: %s", msg)
                await self._save_error_event(msg)
            elif isinstance(msg, ProgressEvent):
                logger.debug("saving progress: %s", msg)
                await self._save_progress_event(msg)
            elif isinstance(msg, CancelledEvent):
                logger.debug("saving cancellation: %s", msg)
                await self._save_cancelled_event(msg)
            else:
                raise TypeError(f"unexpected message type {msg.__class__}")

    @final
    async def _save_result_event(self, result: ResultEvent):
        await self.save_result(result)
        task = await self.get_task(result.task_id)
        task = task.as_resolved(result)
        await self.save_task(task)

    @final
    async def _save_error_event(self, error: ErrorEvent):
        # Update the task retries count
        task = await self.get_task(error.task_id)
        task = task.as_resolved(error)
        await self.save_error(error)
        if task.state is TaskState.QUEUED:
            await self.requeue(task)
        await self.save_task(task)

    @abstractmethod
    async def cancel(self, task_id: str, *, requeue: bool): ...

    @abstractmethod
    async def shutdown_workers(self): ...

    @abstractmethod
    async def _consume(self) -> ManagerEvent: ...

    @abstractmethod
    async def _enqueue(self, task: Task) -> Task: ...

    @abstractmethod
    async def _requeue(self, task: Task): ...

    @abstractmethod
    async def get_health(self) -> dict[str, bool]: ...



import uuid
from typing import Any

from aiohttp import BasicAuth

from icij_worker import Task, TaskError, TaskState
from icij_worker.exceptions import UnknownTask
from icij_worker.utils.http import AiohttpClient


class TaskClient(AiohttpClient):
    def __init__(
        self,
        base_url: str,
        auth: BasicAuth | None = None,
        headers: dict | None = None,
        api_prefix: str = "/api",
    ):
        super().__init__(base_url, auth, headers)
        self._api_prefix = api_prefix

    async def create_task(
        self,
        name: str,
        args: dict[str, Any],
        *,
        id_: str | None = None,
        group: str | None = None,
    ) -> str:
        if id_ is None:
            id_ = _generate_task_id(name)
        task = Task.create(task_id=id_, task_name=name, args=args)
        task = task.model_dump(exclude_unset=True)
        task.pop("created_at")
        url = f"{self._api_prefix}/tasks/{id_}"
        if group is not None:
            if not isinstance(group, str):
                raise TypeError(f"expected group to be a string found {group}")
            url += f"?group={group}"
        async with self._put(url, json=task) as res:
            task_res = await res.json()
        return task_res["id"]

    async def get_task(self, id_: str) -> Task:
        url = f"{self._api_prefix}/tasks/{id_}"
        async with self._get(url) as res:
            task = await res.json()
        if task is None:
            raise UnknownTask(id_)
        task = Task.model_validate(task)
        return task

    async def get_tasks(self) -> list[Task]:
        url = f"{self._api_prefix}/tasks"
        async with self._get(url) as res:
            tasks = await res.json()
        tasks = [Task.model_validate(task) for task in tasks]
        return tasks

    async def get_task_state(self, id_: str) -> TaskState:
        return (await self.get_task(id_)).state

    async def get_task_result(self, id_: str) -> Any:
        url = f"{self._api_prefix}/tasks/{id_}/results"
        async with self._get(url) as res:
            task_res = await res.json()
        return task_res

    async def get_task_error(self, id_: str) -> list[TaskError]:
        url = f"{self._api_prefix}/tasks/{id_}/errors"
        async with self._get(url) as res:
            errors = await res.json()
        return [TaskError.model_validate(e) for e in errors]

    async def delete(self, id_: str):
        url = f"{self._api_prefix}/tasks/{id_}"
        async with self._delete(url):
            pass

    async def delete_all_tasks(self):
        for t in await self.get_tasks():
            await self.delete(t.id)


def _generate_task_id(task_name: str) -> str:
    return f"{task_name}-{uuid.uuid4()}"



from functools import cached_property
from pydantic import BaseModel

from icij_common.pydantic_utils import icij_config


class PostgresConnectionInfo(BaseModel):
    model_config = icij_config()

    host: str = "127.0.0.1"
    password: str = "changeme"
    port: int = 5432
    use_ssl: bool = False
    user: str = "postgres"
    connect_timeout_s: float = 2.0

    def url(self, db: str = "postgres") -> str:
        url = f"postgres://{self.user}:{self.password}@{self.host}:{self.port}/{db}"
        if not self.use_ssl:
            url += "?sslmode=disable"
        return url

    @cached_property
    def kwargs(self) -> dict:
        kwargs = self.model_dump()
        kwargs.pop("use_ssl")
        kwargs["connect_timeout"] = int(kwargs.pop("connect_timeout_s"))
        return kwargs



import logging
from contextlib import asynccontextmanager
from copy import deepcopy
from typing import Any, AsyncContextManager

from aiohttp import BasicAuth, ClientResponse, ClientResponseError, ClientSession
from aiohttp.typedefs import StrOrURL
from typing_extensions import Unpack

logger = logging.getLogger(__name__)


class AiohttpClient(AsyncContextManager):
    def __init__(
        self,
        base_url: str,
        auth: BasicAuth | None = None,
        headers: dict | None = None,
    ):
        self._base_url = base_url
        self._auth = auth
        self._session: ClientSession | None = None
        if headers is None:
            headers = dict()
        self._headers = headers

    async def __aenter__(self):
        self._session = ClientSession(self._base_url, auth=self._auth)
        await self._session.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        await self._session.__aexit__(exc_type, exc_value, traceback)

    @asynccontextmanager
    async def _put(self, url: StrOrURL, *, data: Any = None, **kwargs: Any):
        headers = deepcopy(self._headers)
        headers.update(kwargs.pop("headers", dict()))
        async with self._session.put(url, data=data, headers=headers, **kwargs) as res:
            _raise_for_status(res)
            yield res

    @asynccontextmanager
    async def _get(self, url: StrOrURL, *, allow_redirects: bool = True, **kwargs: Any):
        headers = deepcopy(self._headers)
        headers.update(kwargs.pop("headers", dict()))
        async with self._session.get(
            url, headers=headers, allow_redirects=allow_redirects, **kwargs
        ) as res:
            _raise_for_status(res)
            yield res

    @asynccontextmanager
    async def _delete(self, url: StrOrURL, **kwargs: Unpack["_RequestOptions"]):
        headers = deepcopy(self._headers)
        headers.update(kwargs.pop("headers", dict()))
        async with self._session.delete(url, headers=headers, **kwargs) as res:
            _raise_for_status(res)
            yield res


def _raise_for_status(res: ClientResponse):
    try:
        res.raise_for_status()
    except ClientResponseError as e:
        msg = "request to %s, failed with reason %s"
        logger.exception(msg, res.request_info, res.reason)
        raise e
